class FaspSession < ActiveRecord::Base
  extend ActiveSupport::Memoizable

  has_many   :files,    class_name: "FaspFile",     primary_key: :session_id,
                                                    foreign_key: :session_id,
                                                    dependent: :destroy
  belongs_to :stats,    class_name: "FaspStat",     primary_key: "id",
                                                    dependent: :destroy
  belongs_to :transfer, class_name: "FaspTransfer", primary_key: "transfer_id"

  named_scope :unprocessed, :conditions => { :processing_complete => false }

  #----------------------------------
  # START
  # called by faspex_db_background
  #----------------------------------
  def process
    # all non-relay fasp sessions should have a locally generated transfer token
    mark_as_processed and return unless transfer_request || relay_session?
    es = ESession.find_by_central_id(session_id) || ESession.new
    if es.id != nil
      old_status = es.status
      new_status = FaspSession.to_e_session_status(stats.status)
      # Ignore session updates if it timed out in the past and is still attempting to validate (dest. node hung)
      mark_as_processed and return if (old_status == ESession::STATUS_VALIDATION_TIMED_OUT) and (new_status == ESession::STATUS_VALIDATING)
    end

    es.update_attributes(to_e_session_hash)
    pkg = es.e_package
    pkg.update_from_session(es) unless pkg.nil?

    if EConfiguration.otfv_enabled?
      FaspFile.find_all_by_session_id(session_id).select do |file|
        if file.status == "ERROR"
          err_desc = file.err_desc
          if err_desc and err_desc.starts_with?("VALIDATION FAILED: ")
            pkg.validation_status = EPackage::VALIDATION_FAILED
            pkg.save
            break
          end
        end
      end
    end

    if relay_session?
      relay = PackageRelay.find_by_cookie(transfer_cookie)
      relay.update_from_session(es) if relay
    end
  ensure
    mark_as_processed if done?
  end

  #----------------------------------
  # called by faspex_db_background
  # END
  #----------------------------------

  # Whether this fasp session is for a package relay.
  def relay_session?
    transfer_cookie && transfer_cookie.is_relay?
  end

  def create_from_remote?
    transfer_cookie && transfer_cookie.is_create?
  end

  def mark_as_processed
    self.processing_complete = true
    save
  end

  def soap_address
    server_hostname + "/services/soap/FASPSessionNET-200601"
  end

  def done?
    [
      ESession::STATUS_COMPLETE,
      ESession::STATUS_ERROR,
      ESession::STATUS_STOPPED,
      ESession::STATUS_WILLRETRY
    ].include?(FaspSession.to_e_session_status(transfer.stats.status))
  end

  # cookie is a raw string and transfer_cookie is an instance of TransferCookie that wraps the string cookie
  def transfer_cookie
    @transfer_cookie ||= TransferCookie.new(transfer.cookie)
  rescue InvalidTransferCookie
    logger.warn("Unrecognized cookie when processing fasp_session with id #{id}. #{$!}")
    raise
  end

  def transfer_request
    return @transfer_request if defined? @transfer_request
    return nil unless transfer.cookie.present?
    @transfer_request = TransferRequest.find_by_uuid(transfer_cookie.try(:transfer_uuid))
  end

  def initiator_name
    if transfer_request
      transfer_request.user.try(:name)
    else
      nil
    end
  end

  def self.to_e_session_status(status)
    status.capitalize!
    case status
    when /running/i             then ESession::STATUS_TRANSFERRING
    when /awaiting validation/i then ESession::STATUS_TO_BE_VALIDATED
    when /validating/i          then ESession::STATUS_VALIDATING
    when /cancelled/i           then ESession::STATUS_STOPPED
    when /orphaned/i            then ESession::STATUS_ERROR
    else status
    end
  end

private

  def server_hostname
    return transfer.server_host_name if transfer.server_host_name.present?

    share = Share.find_by_id(transfer_request.share_id) if transfer_request.present?
    return share.node.fallback_primary_address if share.present?

    fasp_node = FaspNode.find_by_node_id(server_node_id)

    fasp_node.present? ? fasp_node.node.fallback_primary_address : 'Unknown'
  end

  def to_e_session_hash
    e_session_hash = {
      :e_session_start_notification_time  => utc_time_for("started_at"),
      :e_session_stop_notification_time   => utc_time_for("stopped_at"),
      :status                             => FaspSession.to_e_session_status(stats.status),
      :central_id                         => session_id,
      :e_package_id                       => transfer_request.e_package_id,
      :token_uuid                         => transfer.token,
      :initiator_user_id                  => transfer_request.user_id,
      :initiator_name                     => initiator_name,
      :system_user_name                   => transfer.user,
      :remote_host                        => client_ip,

      :direction                          => transfer_cookie.type,

      :error_code                         => stats.err_code.to_i,
      :error_description                  => stats.err_desc,

      :bytes_written                      => stats.bytes_written,
      :bytes_transferred                  => stats.bytes_transferred,
      :bytes_pretransfer                  => stats.bytes_pretransfer,

      :files_completed                    => stats.files_complete,

      :files_failed                       => stats.files_failed,
      :files_pretransfer                  => stats.files_pretransfer,

      :seconds_elapsed                    => stats.elapsed_usec.to_i / 1_000_000.0,
      :partial_package                    => transfer_request.partial_package,

      :share_id                           => transfer_request.share_id,
      :delivery_id                        => transfer_request.delivery_id,
      :transfer_request                   => transfer_request,

      :soap_uri                           => soap_address
    }
    e_session_hash
  end

  # mysql does not allow time zones for datetimes. Fasp always reports in UTC but we want
  # to display a time according to the server's time zone. This forces the time from fasp
  # to be parsed as UTC.
  def utc_time_for(attribute_name)
    time = transfer.stats.attributes_before_type_cast[attribute_name]
    Time.parse(time + " +0000") if time
  end

  # return address of transfer server node that handled this session, as reported by node poller
  def local_address
    fasp_node = FaspNode.find_by_node_id(client_node_id, :order => "reported_at DESC")
    fasp_node ? fasp_node.host_ip : ''
  end
end
