class FaspexConfig::TemplateUser < FaspexConfig; end

class TemplateUser < User
  include User::Permissionable
  before_validation_on_create :ensure_one_record
  before_destroy do
    logger.error "Template user cannot be destroyed, stacktrace: #{caller}"
    return false
  end

  TIME_RELATIVE_ATTRIBUTES = [:days_til_expiration, :days_til_destruction]
  COPYABLE_ATTRIBUTES = [*User::Permissionable::ATTRIBUTES, :account_expires, :self_destructs]

  attr_accessible *COPYABLE_ATTRIBUTES, *TIME_RELATIVE_ATTRIBUTES
  attr_accessor *TIME_RELATIVE_ATTRIBUTES

  named_scope :users_with_default_password_policy, :conditions => 'not override_default_password_policy'

  named_scope :users_with_password_expiration_date_after, lambda { |ed|
    { :conditions => ['password_expiration_date IS NULL OR password_expiration_date > :ed', {:ed => ed} ] }
  }

  validates_numericality_of :days_til_expiration,
    only_integer: true,
    greater_than: 0,
    less_than_or_equal_to: 730,
    if: :account_expires

  validates_numericality_of :days_til_destruction,
    only_integer: true,
    greater_than: 0,
    less_than_or_equal_to: 730,
    if: :self_destructs


  def ensure_one_record
    record = TemplateUser.first
    unless record.nil?
      errors.add_to_base "TemplateUser record already exists, cannot create another"
      false
    end
  end

  # Returns those attributes only which the TemplateUser can actually assign
  def copyable_attributes
    attributes.symbolize_keys.slice(*COPYABLE_ATTRIBUTES).
      merge(expiration_date: expiration_date, destruction_date: destruction_date)
  end

  def expiration_date;  Time.now + TemplateUser.first.days_til_expiration.days;  end
  def destruction_date; Time.now + TemplateUser.first.days_til_destruction.days; end

  TIME_RELATIVE_ATTRIBUTES.each do |meth|
    define_method(meth) { FaspexConfig::TemplateUser[meth] || 90 }                         # def days_til_expiration; FaspexConfig[meth] || 90 end
    define_method(meth.to_s + '=') { |days| FaspexConfig::TemplateUser[meth] = days.to_i } # def days_til_expiration= days; FaspexConfig::TemplateUser[:day_til_expiration = days.to_i end
    define_method(meth.to_s + '_before_type_cast') { send meth }                           # def days_til_expiration_before_type_cast; self.days_til_expiration end
  end

  def self.copyable_attributes
    TemplateUser.first.copyable_attributes
  end

  def self.workgroups
    TemplateUser.first.workgroups
  end

end
