require_dependency 'packages/package_methods'
require 'builder'

class User

  # All cases ending in ".atom" are used externally in the atom feed and should never be called from
  # within the Faspex application.
  def deliveries(box, page=1, count=nil, startIndex=nil)
    case box
    when 'inbox.atom' then atom_recipient_deliveries(page, count, startIndex)
    when 'inbox'   then inbox_deliveries(page, count, startIndex)
    when 'archive.atom' then inbox_deliveries(page, count, startIndex, true)
    when 'sent.atom'    then sent_deliveries(page, count, startIndex)
    else raise ArgumentError, 'Invalid box'
    end
  end

  def num_per_page
    10
  end
  alias_method :feed_entries_per_page, :num_per_page

  def order
    'deliveries.sequence_id DESC'
  end

  # Returns all relevant recipient_deliveries for an inbox.atom feed, replaces inbox_deliveries() for the inbox.atom feed
  # @param page [String, #paginate_by_sql] adds LIMIT and OFFSET to sql query to limit results by page
  # @param count [String, #to_i]
  # @param startIndex [String, #to_i ] starting index for the mysql query for all RecipientDelivery objs
  # @param show_archived [boolean] check to see if archived packages should be included
  # @return [RecipientDelivery] objects that are fed into the /delivery/views/_content.xml.builder
  def atom_recipient_deliveries(page, count=nil, startIndex=nil, show_archived = false)
    recipient_id_conditions = "(recipient_id=#{id} AND recipient_type='User')"
    workgroup_ids = connection.select_values("SELECT DISTINCT e_workgroup_id FROM e_memberships WHERE user_id=#{id} AND can_read=1")
    unless workgroup_ids.empty?
      recipient_id_conditions += " OR (recipient_id IN (#{workgroup_ids.join(",")}) AND recipient_type='EWorkgroup')"
    end

    sql = "SELECT DISTINCT rd.* FROM recipient_deliveries rd USE INDEX (idx_recipient_id_recipient_type)
            STRAIGHT_JOIN deliveries d ON d.id=rd.delivery_id
            STRAIGHT_JOIN e_packages p ON d.e_package_id=p.id
            WHERE
            (#{recipient_id_conditions})
            AND rd.package_deleted = 0
            AND #{sql_archived_conditions(show_archived)}
            AND d.state = '#{Delivery::STATE_RELEASED}'
            AND p.e_package_contents_exist = 1
            AND p.e_package_files_deleted_by_admin_time IS NULL
            ORDER BY d.sequence_id DESC"

    count = count.to_i
    count = num_per_page if count < 1

    if startIndex.present?
      startIndex = startIndex.to_i
      startIndex = 0 if startIndex < 1

      # Limit <start> <num>
      sql += " LIMIT #{startIndex.to_i}, #{count.to_i}"
      RecipientDelivery.find_by_sql(sql)
    else
      RecipientDelivery.paginate_by_sql(
          sql,
          {
              :page => page,
              :per_page => count,
          }
      )
    end

  end


  def inbox_deliveries(page, count=nil, startIndex=nil, show_archived = false)
    recipient_id_conditions = "(recipient_id=#{id} AND recipient_type='User')"
    workgroup_ids = connection.select_values("SELECT DISTINCT e_workgroup_id FROM e_memberships WHERE user_id=#{id} AND can_read=1")
    unless workgroup_ids.empty?
      recipient_id_conditions += " OR (recipient_id IN (#{workgroup_ids.join(",")}) AND recipient_type='EWorkgroup')"
    end

    sql = "SELECT DISTINCT d.* FROM recipient_deliveries rd USE INDEX (idx_recipient_id_recipient_type)
            STRAIGHT_JOIN deliveries d ON d.id=rd.delivery_id
            STRAIGHT_JOIN e_packages p ON d.e_package_id=p.id
            WHERE
            (#{recipient_id_conditions})
            AND rd.package_deleted = 0
            AND #{sql_archived_conditions(show_archived)}
            AND d.state = '#{Delivery::STATE_RELEASED}'
            AND p.e_package_contents_exist = 1
            AND p.e_package_files_deleted_by_admin_time IS NULL
            ORDER BY d.sequence_id DESC"

    count = count.to_i
    count = num_per_page if count < 1

    if startIndex.present?
      startIndex = startIndex.to_i
      startIndex = 0 if startIndex < 1

      # Limit <start> <num>
      sql += " LIMIT #{startIndex.to_i}, #{count.to_i}"
      Delivery.find_by_sql(sql)
    else
      Delivery.paginate_by_sql(
          sql,
          {
              :page => page,
              :per_page => count,
          }
      )
    end

  end

  def sent_deliveries(page, count=nil, startIndex=nil, show_archived = false)
    count = count.to_i
    count = num_per_page if count < 1

    if startIndex.present?
      startIndex = startIndex.to_i
      startIndex = 0 if startIndex < 1

      sql = "SELECT DISTINCT d.* FROM recipient_deliveries rd USE INDEX (idx_recipient_id_recipient_type)
              STRAIGHT_JOIN deliveries d ON d.id=rd.delivery_id
              STRAIGHT_JOIN e_packages p ON d.e_package_id=p.id
              WHERE
              d.sender_id = #{id}
              AND rd.package_deleted = 0
              AND rd.archived = #{show_archived ? 1 : 0}
              AND d.sequence_id IS NOT NULL
              AND p.e_package_contents_exist = 1
              AND p.e_package_files_deleted_by_admin_time IS NULL
              AND p.e_package_upload_status = 'completed'
              ORDER BY d.sequence_id DESC
              LIMIT #{startIndex.to_i}, #{count.to_i}"
      Delivery.find_by_sql(sql)
    else
      Delivery.paginate(
          :page => page,
          :per_page => count,
          :include => [:e_package, :recipient_deliveries],
          :conditions => sent_conditions(show_archived),
          :order => order
      )
    end
  end

  def sent_conditions(show_archived)
    conditions =
        package_exists_conditions + ' AND ' +
        'deliveries.sender_id = ?' + ' AND ' +
        'deliveries.state = ?'
    unless show_archived
      conditions = conditions + ' AND ' + hide_archived_conditions
    end
    [conditions, id, Delivery::STATE_RELEASED]
  end

  def inbox_conditions(show_archived)
    conditions =
        package_exists_conditions + ' AND ' +
        '((recipient_deliveries.recipient_type = "User" AND ' +
        'recipient_deliveries.recipient_id = ?) OR ' +
        '(recipient_deliveries.recipient_type = "EWorkgroup" AND '+
        'e_memberships.user_id = ? AND ' +
        'e_memberships.can_read = 1))'
    unless show_archived
      conditions = conditions + ' AND ' + hide_archived_conditions
    end
    [conditions, id, id]
  end

  def package_exists_conditions
    'e_packages.e_package_contents_exist = 1 AND ' +
        'e_packages.e_package_files_deleted_by_admin_time IS NULL AND ' +
        'e_packages.e_package_upload_status = "Completed" AND ' +
        'deliveries.sequence_id IS NOT NULL'
  end

  def sql_archived_conditions(show_archived)
    if show_archived
      'rd.archived=1'
    else
      "rd.archived=0 AND p.e_package_status <> '#{EPackage::PENDING_DELETION}'"
    end
  end

  def hide_archived_conditions
    'recipient_deliveries.package_deleted = 0 AND ' +
      'recipient_deliveries.archived = 0 AND ' +
      "e_packages.e_package_status <> '#{EPackage::PENDING_DELETION}'"
  end

  def normalized_address_for_recipient(recipient)
    recipient.strip!
    is_workgroup = (recipient[0,1] == '*')
    recipient = recipient[1..-1] if is_workgroup
    if recipient.include?("@") && !is_workgroup
      return recipient
    else
      faspex_host = EConfiguration.get_server_address.downcase
      return "#{CGI::escape(recipient)}@#{faspex_host}"
    end
  end

  def normalized_name_for_recipient(name)
    name.strip!
    name = name[1..-1] if name[0,1] == '*'
    return name.strip
  end

  def recipients_for_delivery(delivery)
    recipients = []
    delivery.all_recipients.each do |recipient|
      recipients << {:name => normalized_name_for_recipient(recipient), :email => normalized_address_for_recipient(recipient.strip)}
    end
    return recipients
  end

end
