#
# @mailbox_item has two possible types: RecipientDelivery or Delivery
#
# We use the @mailbox_item to get the transfer spec for package download.
#
# RecipientDelivery
#
#    This is when the current_user is looking at their inbox. We show all of
#    their received packages (recorded as RecipientDeliveries).
#
#    We want the user to download from their assigned share. The
#    RecipientDelivery (either a UserDelivery or WorkgroupDelivery)
#    contains the storage_share where the user should download from.
#
# Delivery
#
#    This is when the current_user is looking at their outbox, we show all of
#    the packages they sent (recorded as Deliveries).
#
#    We have the current_user download from the package's storage_share
#    (@delivery.e_package). The assumption is that this is the
#    current_user's assigned share (since they were the sender) and is
#    hence where they should download from.

recipient_delivery = (@delivery.is_a? RecipientDelivery) ? @delivery : nil # If recipient_delivery object received, store it
@delivery = (@delivery.is_a? RecipientDelivery) ? Delivery.find_by_id(@delivery.delivery_id) : @delivery.delivery # If recipient_delivery object received, get it's corresponding Delivery object
package = @delivery.e_package
status = package.package_status_for_inbox_feed
unopened = !package.user_viewed?(current_user.id)

xml.title(@delivery.title)
xml.content(@delivery.note)
@delivery.metadata_xml(:builder => xml)
xml.id(package.e_package_uuid)

xml.sequence_id(@delivery.sequence_id)
xml.published(@delivery.created_at.utc.strftime("%Y-%m-%dT%H:%M:%SZ"))
xml.updated(@delivery.updated_at.utc.strftime("%Y-%m-%dT%H:%M:%SZ"))
if @delivery.e_package.completed?
  xml.completed(@delivery.completed_at.utc.strftime("%Y-%m-%dT%H:%M:%SZ"))
else
  xml.completed
end
xml.tag!('package:property', 'type' => status) if status && package.toplevel_items.empty?
xml.tag!('package:property', 'type' => 'unopened') if unopened
xml.author do
  xml.name(@delivery.sender_name)
  xml.email(@delivery.sender.try(:email))
end
@delivery.recipient_deliveries.public.each do |rd|
  xml.tag!('package:to') do
    xml.tag!('package:name', rd.recipient_name)
    if (rd.recipient && rd.recipient.respond_to?(:email))
      xml.tag!('package:email', rd.recipient.email)
    end
    xml.tag!('package:recipient_delivery_id', rd.id)
  end
end

# This will always return the actual Delivery id, regardless of whether
# @delivery is a Delivery or RecipientDelivery
xml.tag!('package:parent_delivery_id', @delivery.delivery.id)

#SDK_TEAM_APPLE_CHANGE
xml.tag!('package:delivery_id', @delivery.id)
if current_user == @delivery.sender
  @delivery.recipient_deliveries.private.each do |rd|
    xml.tag!('package:to_private') do
      xml.tag!('package:name', rd.recipient_name)
      if (rd.recipient && rd.recipient.respond_to?(:email))
        xml.tag!('package:email', rd.recipient.email)
      end
    end
  end
end
if package.inner_directory_exists? && !validation_failed?(@delivery) && current_user.downloads_allowed?
  include_token = current_user.is_a?(ExternalUser)

  # Gets the request IP for all default inbox downloads, otherwise chooses host of storage_share node
  if recipient_delivery.nil? # Default download
    host = @delivery.remote_host(request.remote_ip, request.host)
  else
    host = (recipient_delivery.default_download_location.node.primary_address.empty? ? recipient_delivery.default_download_location.node.host : recipient_delivery.default_download_location.node.primary_address)
    # Download from final storage share
  end
  storage_share = recipient_delivery.nil? ? nil : recipient_delivery.storage_share.id
  package_transfer_spec = nil
  original_path = nil
  begin
    package_transfer_spec = @delivery.get_transfer_spec(
      ESession::DIRECTION_DOWNLOAD, ['/'], current_user,
        {
          :token => include_token,
          :host => host,
          :share => storage_share
        }
    )

    package_transfer_spec.add_share_to_tags(storage_share) unless storage_share.nil?

    xml.link(
      'rel' => 'package',
      'href' => package_transfer_spec.to_fasp_url.to_s
    )
    original_path = package_transfer_spec.paths.first['source'] if package_transfer_spec.present?
  rescue
    xml.tag!('package:transfer_spec_error', 'Failed to generate FASP url for package')
  end

  package.toplevel_items.each do |item|
    raise unless package_transfer_spec || include_token

    item_transfer_spec = if include_token
                           @delivery.get_transfer_spec(
                             ESession::DIRECTION_DOWNLOAD, [item.basename], current_user,
                              {
                                :token => true,
                                :host => host,
                                :share => storage_share
                              }
                             )
                         else
                           package_transfer_spec.paths.first['source'] = File.join(original_path, item.basename)
                           package_transfer_spec
                         end

    item_transfer_spec.add_share_to_tags(storage_share) unless storage_share.nil?

    begin
      xml.link(
        'rel' => 'enclosure',
        'href' => item_transfer_spec.to_fasp_url.to_s,
        'size' => item.size,
        'type' => item.type
      )
    rescue
      xml.tag!('package:transfer_spec_error', "Failed to generate FASP url for package item: #{item.basename}")
    end
  end
end
